/*
 * fr32_fir_test.c
 *
 *  Created on: 2013/05/02
 *      Author: takemasa
 */


#include "fr32_fir_test.h"
#include <limits.h>
#include <stdio.h>


typedef long long int true_fract32;

#define SATP(x)  (((x)>2147483647LL)?(2147483647LL):x)
#define SAT(x) (((x)>=-2147483648LL)?(SATP(x)):-2147483648LL)

#define _mult_fr1x32x32(x,y) SAT(((true_fract32)(x)*(true_fract32)(y)+0x40000000)>>31)

#define _sub_fr1x32(x,y) SAT((true_fract32)(x)-(y))

#define _add_fr1x32(x,y) SAT((true_fract32)(x)+(y))


void clearBuffer ( fract32 buf[], int count)
{
    int i;
    for ( i= 0; i<count; i++ )
        buf[i] = 0;
}

/*
 * Basic test to see the continuity of HH and HL&LH product.
 */
#define TAPS_01 2
#define NUMSAMPLE_01 4

fract32 coeff_01[TAPS_01]=
    {
            0x08000000,     // >>4
            0x0FFFF000,     // pattern to use both H&L
    };

fract32 input_01[NUMSAMPLE_01] =
    {
        0x0FFFF000,         // pattern to use both H&L
        0,
        0x00800000,         // >>16
    };

fract32 desired_01[NUMSAMPLE_01] =
    {
        0x00FFFF00,         // 0x0FFFF000 >> 4
        0x01FFFC00,         // 0x0FFFF000 * 0x0FFF000 ( signed fixed )
        0x00080000,         // 0x08 >> 16
        0x000FFFF0,         // 0x0FFFF000 >>16
    };

void test_01_basic()
{
    fract32 dellin[TAPS_01];    // internal variable of fir filter
    TFIR32 filter;              // filter object
    fract32 output[NUMSAMPLE_01];
    int i;


        // clear output buffer
    clearBuffer( output, NUMSAMPLE_01);
        // create a filter object
    fr32_fir_init( &filter, TAPS_01, coeff_01, dellin, 1);
        // apply filter
    fr32_fir( &filter, input_01, output, NUMSAMPLE_01);

    for ( i=0; i<NUMSAMPLE_01; i++)
    {
        if ( output[i] != desired_01[i] )
        {
            printf( "test01 NG :output[%2d] = 0x%08X but should be 0x%08X\n", i, output[i], desired_01[i] );
            return;
        }
    }
    printf ("test01 OK\n");
}

#undef TAPS_01
#undef NUMSAMPLE_01

/*
 * Sharrow Round up test for H*L+L*M. This test certify the round up from the bit under LSB.
 */
#define TAPS_02 2
#define NUMSAMPLE_02 5

fract32 coeff_02[TAPS_02]=
    {
            0x08000000,     // >>4
            0x00000000,
    };

fract32 input_02[NUMSAMPLE_02] =
    {
        0x00000010,
        0x00000008,
        0x00000004,
        0x00000002,
        0x00000001,
    };

fract32 desired_02[NUMSAMPLE_02] =
    {
            0x00000001,     // 0x010 >> 4
            0x00000001,     // this must be 1 because or round up.
            0x00000000,     // 0
            0x00000000,
            0x00000000,
    };


void test_02_round_near_1()
{
    fract32 dellin[TAPS_02];    // internal variable of fir filter
    TFIR32 filter;              // filter object
    fract32 output[NUMSAMPLE_02];
    int i;


        // clear output buffer
    clearBuffer( output, NUMSAMPLE_02);
        // create a filter object
    fr32_fir_init( &filter, TAPS_02, coeff_02, dellin, 1);
        // apply filter
    fr32_fir( &filter, input_02, output, NUMSAMPLE_02);

    for ( i=0; i<NUMSAMPLE_02; i++)
    {
        if ( output[i] != desired_02[i] )
        {
            printf( "test02 NG :output[%2d] = 0x%08X but should be 0x%08X\n", i, output[i], desired_02[i] );
            return;
        }
    }
    printf ("test02 OK\n");
}

#undef TAPS_02
#undef NUMSAMPLE_02


/*
 * L*L multiplying test.
 */
#define TAPS_03 2
#define NUMSAMPLE_03 5

fract32 coeff_03[TAPS_03]=
    {
            0x00008000,
            0x00000000,
    };

fract32 input_03[NUMSAMPLE_03] =
    {
        0x00020000,
        0x00010000,
        0x00008000,
        0x00004000,
        0x00000000,
    };

fract32 desired_03[NUMSAMPLE_03] =
    {
            0x00000002,     // H*L
            0x00000001,     // H*L
            0x00000001,     // L*L but round
            0x00000000,     // L*L
            0x00000000,
    };


void test_03_L_by_L()
{
    fract32 dellin[TAPS_03];    // internal variable of fir filter
    TFIR32 filter;              // filter object
    fract32 output[NUMSAMPLE_03];
    int i;


        // clear output buffer
    clearBuffer( output, NUMSAMPLE_03);
        // create a filter object
    fr32_fir_init( &filter, TAPS_03, coeff_03, dellin, 1);
        // apply filter
    fr32_fir( &filter, input_03, output, NUMSAMPLE_03);

    for ( i=0; i<NUMSAMPLE_03; i++)
    {
        if ( output[i] != desired_03[i] )
        {
            printf( "test03 NG :output[%2d] = 0x%08X but should be 0x%08X\n", i, output[i], desired_03[i] );
            return;
        }
    }
    printf ("test03 OK\n");
}

#undef TAPS_03
#undef NUMSAMPLE_03

/*
 * Negative value and saturation test.
 */
#define TAPS_04 2
#define NUMSAMPLE_04 6

fract32 coeff_04[TAPS_04]=
    {
        0x80000000,
        0x7FFFFFFF
    };

fract32 input_04[NUMSAMPLE_04] =
    {
        0x40000000,
        0x00000000,
        0x00000002,
        0x00000000,
        0x7FFFFFFF,
        0x80000000
    };

fract32 desired_04[NUMSAMPLE_04] =
    {
        0xC0000000,     // -0.5
        0x40000000,     //
        0xFFFFFFFE,     // -2LSB
        0x00000002,
        0x80000001,     // -1.0+1LSB
        0x7FFFFFFF,     // 1.0-1LSB ( Saturated )
    };


void test_04_netagive()
{
    fract32 dellin[TAPS_04];    // internal variable of fir filter
    TFIR32 filter;              // filter object
    fract32 output[NUMSAMPLE_04];
    int i;


        // clear output buffer
    clearBuffer( output, NUMSAMPLE_04);
        // create a filter object
    fr32_fir_init( &filter, TAPS_04, coeff_04, dellin, 1);
        // apply filter
    fr32_fir( &filter, input_04, output, NUMSAMPLE_04);

    for ( i=0; i<NUMSAMPLE_04; i++)
    {
        if ( output[i] != desired_04[i] )
        {
            printf( "test04 NG :output[%2d] = 0x%08X but should be 0x%08X\n", i, output[i], desired_04[i] );
            return;
        }
    }
    printf ("test04 OK\n");
}

#undef TAPS_04
#undef NUMSAMPLE_04

/*
 * Test the TAPs and input length handling.
 */
#define TAPS_05 4
#define NUMSAMPLE_05 10

fract32 coeff_05[TAPS_05]=
    {
        0x08000000,
        0x08000000,
        0x08000000,
        0x08000000,
    };

fract32 input_05[NUMSAMPLE_05] =
    {
        0x08000000,     //
        0x04000000,     //
        0x02000000,     //
        0x01000000,     //
        0x00800000,     //
        0x00400000,     //
        0x00200000,
        0x00100000,
        0x00080000,
        0x00040000
    };

fract32 desired_05[NUMSAMPLE_05] =
    {
        0x00800000,     //
        0x00C00000,     //
        0x00600000,     //
        0x00300000,     //
        0x00180000,     //
        0x000C0000,     //
        0x00000000,
        0x00000000,     //
        0x00000000,     //
        0x00000000,     //
    };


void test_05_length()
{
    fract32 dellin[TAPS_05];    // internal variable of fir filter
    TFIR32 filter;              // filter object
    fract32 output[NUMSAMPLE_05];
    int i;


        // clear output buffer
    clearBuffer( output, NUMSAMPLE_05);
        // create a filter object
    fr32_fir_init( &filter, 2, coeff_05, dellin, 1);
        // apply filter
    fr32_fir( &filter, input_05, output, 6);

    for ( i=0; i<NUMSAMPLE_05; i++)
    {
        if ( output[i] != desired_05[i] )
        {
            printf( "test05 NG :output[%2d] = 0x%08X but should be 0x%08X\n", i, output[i], desired_05[i] );
            return;
        }
    }
    printf ("test05 OK\n");
}

#undef TAPS_05
#undef NUMSAMPLE_05

/* ------------------------------ Interpolator ---------------------------- */

/*
 * Basoc interpolation test.
 */
#define TAPS_06 12
#define NUMSAMPLE_06 12
#define RATIO_06 4

fract32 coeff_06[TAPS_06]=
    {
        0x01000000,
        0x02000000,
        0x03000000,
        0x04000000,
        0x05000000,
        0x06000000,
        0x07000000,
        0x08000000,
        0x0A000000,
        0x0B000000,
        0x0C000000,
        0x0D000000,

    };

fract32 input_06[NUMSAMPLE_06] =
    {
        0x08000000,
        0x00800000,
        0x00080000,
        0x00008000,
        0x00000800,
        0x00000080,
        0x00000008,
    };

fract32 desired_06[NUMSAMPLE_06*RATIO_06];


void test_06_basic_interpolate()
{
    fract32 dellin[TAPS_06];    // internal variable of fir filter
    TFIR32 filter;              // filter object
    fract32 output[NUMSAMPLE_06*RATIO_06];
    int i, j;


        // clear output buffer
    clearBuffer( output, NUMSAMPLE_06*RATIO_06);
        // create a filter object
    fr32_fir_init( &filter, TAPS_06, coeff_06, dellin, RATIO_06);        // interpolation 3
        // apply filter
    fr32_interpolate( &filter, input_06, output, NUMSAMPLE_06);

        // create reference output
    clearBuffer( desired_06, NUMSAMPLE_06*RATIO_06);
    for ( j=0; j<7; j++)
        for( i=0; i<(NUMSAMPLE_06)*RATIO_06; i++)
            if ( i< TAPS_06 )
                    desired_06[i+j*RATIO_06] += _mult_fr1x32x32(coeff_06[i], input_06[j]);


    for ( i=0; i<NUMSAMPLE_06*RATIO_06; i++)
    {
        if ( output[i] != desired_06[i] )
        {
            printf( "test_06 NG :output[%2d] = 0x%08X but should be 0x%08X\n", i, output[i], desired_06[i] );
            return;
        }
    }
    printf ("test_06 OK\n");
}

#undef TAPS_06
#undef NUMSAMPLE_06


/*
 * interpolation test. compare with fir.
 */
#define TAPS_07 12
#define NUMSAMPLE_07 12
#define RATIO_07 4

fract32 coeff_07[TAPS_07]=
    {
        0x01000000,
        0x02000000,
        0x03000000,
        0x04000000,
        0x05000000,
        0x06000000,
        0x07000000,
        0x08000000,
        0x0A000000,
        0x0B000000,
        0x0C000000,
        0x0D000000,

    };

fract32 input_07[NUMSAMPLE_07] =
    {
        0x08000000,
        0x00800000,
        0x00080000,
        0x00008000,
        0x00000800,
        0x00000080,
        0x00000008,
    };

fract32 desired_07[NUMSAMPLE_07*RATIO_07];


void test_07_compare_interpolate()
{
    fract32 dellin[TAPS_07*RATIO_07];    // internal variable of fir filter
    TFIR32 filter;              // filter object
    fract32 output[NUMSAMPLE_07*RATIO_07];
    fract32 input[NUMSAMPLE_07*RATIO_07];
    int i, j, k;


        // clear output buffer
    clearBuffer( output, NUMSAMPLE_07*RATIO_07);

        // prepare interpolated input
    k=0;
    for ( i=0; i<NUMSAMPLE_07; i++){
        input[k++] = input_07[i];
        for( j=1;j<RATIO_07;j++ )
            input[k++] = 0;
    }

        // create a filter object
    fr32_fir_init( &filter, TAPS_07, coeff_07, dellin, RATIO_07);        // interpolation 3
        // apply filter
    fr32_interpolate( &filter, input_07, desired_07, NUMSAMPLE_07);

        // create reference output
    fr32_fir_init( &filter, TAPS_07, coeff_07, dellin, 1);        // interpolation 3
        // apply filter
    fr32_fir( &filter, input, output, NUMSAMPLE_07*RATIO_07);



    for ( i=0; i<NUMSAMPLE_07*RATIO_07; i++)
    {
        if ( output[i] != desired_07[i] )
        {
            printf( "test_07 NG :output[%2d] = 0x%08X but should be 0x%08X\n", i, output[i], desired_07[i] );
            return;
        }
    }
    printf ("test_07 OK\n");
}

#undef TAPS_07
#undef NUMSAMPLE_07


/*
 * interpolation test. compare with fir.
 */
#define TAPS_08 12
#define NUMSAMPLE_08 40
#define RATIO_08 4

fract32 coeff_08[TAPS_08]=
    {
        0x01000000,
        0x02000000,
        0x03000000,
        0x04000000,
        0x05000000,
        0x06000000,
        0x07000000,
        0x08000000,
        0x0A000000,
        0x0B000000,
        0x0C000000,
        0x0D000000,

    };

fract32 input_08[NUMSAMPLE_08] =
    {
        0x08000000,
        0x00800000,
        0x00080000,
        0x80000000,
        0xAAAAAAAA,
        0x00000080,
        0x00000008,
        0xDEADDEAD,
        0xFFFFFFFF,
        0x7FFFFFFF,
        0x08000000,
        0x00800000,
        0x00080000,
        0x80000000,
        0xAAAAAAAA,
        0x00000080,
        0x00000008,
        0xDEADDEAD,
        0xFFFFFFFF,
        0x7FFFFFFF,
        0x08000000,
        0x00800000,
        0x00080000,
        0x80000000,
        0xAAAAAAAA,
        0x00000080,
        0x00000008,
        0xDEADDEAD,
        0xFFFFFFFF,
        0x7FFFFFFF,
        0x08000000,
        0x00800000,
        0x00080000,
        0x80000000,
        0xAAAAAAAA,
        0x00000080,
        0x00000008,
        0xDEADDEAD,
        0xFFFFFFFF,
        0x7FFFFFFF,
    };

fract32 desired_08[NUMSAMPLE_08];


void test_08_basic_decimate()
{
    fract32 dellin[TAPS_08];    // internal variable of fir filter
    TFIR32 filter;              // filter object
    fract32 output[NUMSAMPLE_08];
    int i, j, k;


        // clear output buffer
    clearBuffer( output, NUMSAMPLE_08);

        // create a filter object
    fr32_fir_init( &filter, TAPS_08, coeff_08, dellin, RATIO_08);        // interpolation 3
        // apply filter
    fr32_decimate( &filter, input_08, output, NUMSAMPLE_08);

        // create reference output
    fr32_fir_init( &filter, TAPS_08, coeff_08, dellin, 1);        // interpolation 3
        // apply filter
    fr32_fir( &filter, input_08, desired_08, NUMSAMPLE_08);

        // prepare interpolated input
    for ( i=0; i<NUMSAMPLE_08/RATIO_08; i++){
        desired_08[i] = desired_08[(i+1)*RATIO_08-1];
    }




    for ( i=0; i<NUMSAMPLE_08/RATIO_08; i++)
    {
        if ( output[i] != desired_08[i] )
        {
            printf( "test_08 NG :output[%2d] = 0x%08X but should be 0x%08X\n", i, output[i], desired_08[i] );
            return;
        }
    }
    printf ("test_08 OK\n");
}

#undef TAPS_08
#undef NUMSAMPLE_08


