/*
* 32bit FIR filter implementation.
* Todo: not tested yet
* function prototype
*	void fr32_interpolate(TFIR32 *filter, fract32 in[], fract32 out[], int count );
*
* parameters
*   FP+20	-	int count
*   FP+16	R2	fract32 * out
*	FP+12	R1	fract32 * in
*	FP+ 8	R0	TFIR32 * filter
*
* return
*	none
*
* side effect
*   out[] : obtain output data
*   filter->head : the head of sample input is upldated.
*
* TFIR32 structure
*   typedef struct
*   {
*		int taps;
*		fract32 * coeff;
*		fract32 * delayline;
*		fract32 * head;
*		int ratio
*		int tapsByRatio;
*	}TFIR32;

* register layout
*   P0 : filter
*   P1 : in
*	P3 : count : loop counter's initial value
*	P4 : taps/ratio : loop counter's initial value
*   I0 : coeff <- points the coeff sample
*   B0 : coeff
*   L0 : taps*4
*	M0 : ratio * 4
*   I1 : head  <- points the delayline head and sample point
*   B1 : delayline
*   L1 : tapsByRatio*4
*	I2 : out
*   R0 : *coeff
*   R1 : *delay
*	R2 : ratio
*	R4 : output sample
*	R5 : taps
*	R6 : bit 15
*	R7 : bit 14
*/

	.text
	.align 4
	.global _fr32_interpolate;
	.type _fr32_interpolate, STT_FUNC;

_fr32_interpolate:
	link	0;
	[--sp] = (r7:4, p5:3);		// save all preserved register

		/* Set up registers */
	p0 = r0;					// pointer to filter
	p1 = r1;					// pointer to in
	i2 = r2;					// pointer to out
	r5 = [p0];					// taps
	r3 = r5 << 2;				// taps*4
	l0 = r3;					// taps*4
	r5 = [p0+20];				// taps/ratio
	p4 = r5;					// taps/ratio
	r3 = r5 << 2;				// taps*4/ratio
	l1 = r3;					// taps*4/ratio
	r7 = [p0+4];				// coeff
	b0 = r7;
	i0 = b0;					// coeff
	r7 = [p0+8];
	b1 = r7;					// delayline
	r7 = [p0+12];
	i1 = r7;					// head
	r2 = [p0+16];				// ratio
	r3 = r2 << 2				// ratio * 4
	m0 = r3;					// ratio *4
	r3 = [fp+20];				// load count
	r3 *= r2;					// count * ratio
	p3 = r3;
	r7 = 0x4000(z);				// bit 14
	p4 += -1;
	r2 = 1;						// so, first execution triggers to get one sample.



		/* outer loop */
	loop count lc0 = p3;
	loop_begin count;

		i0 += 4;					// forward to next phase of poly phase filter
		r2 += -1;					// decrement ratio
		cc = az;					// zero?
		if !cc jump pushToDelayline(bp);	// if not zero, do not push input data to delayline
									// push input sample into delayline
			i0 = b0;					// rewind coeff pointer
			r1 = [p1++];				// load input sample
			r2 = [p0+16];				// road ratio
			i1 -= 4;					// step back the pointer
			[i1] = r1;
pushToDelayline:
									// load input sample.
		a1 = r7 || r1 = [i1++];		// set bit14 as bias to rounding

			/* inner loop first half*/
		a0 = 0 || r0 = [i0++m0];				// load coeff & delaydata
		loop taps_first lc1 = p4;
		loop_begin taps_first;

				// it is important to calc coef at first half.
				// load coeff & delaydata
			a1 += r0.H * r1.L (m), a0 += r0.L * r1.L (fu) || r0 = [i0++m0] || r1 = [i1++];

		loop_end taps_first;
		a1 += r0.H * r1.L (m), a0 += r0.L * r1.L (fu);		// it is important to calc coef at first half.
			/* end of inner loop second half */

		a0 = a0 >> 16;					// scale down the unsigned integer.
		a0 += a1;						// compose intermediate result
		a1 = a0;						// unfortunately, we do not have a0 = a1;

			/* inner loop second half */
		a0 = 0 || r0 = [i0++m0] || r1 = [i1++];				// load coeff & delaydata
		loop taps_sec lc1 = p4;
		loop_begin taps_sec;

				// load coeff & delaydata
			a0 += r1.H * r0.H, a1 += r1.H * r0.L (m) || r0 = [i0++m0] || r1 = [i1++];

		loop_end taps_sec;
		a0 += r1.H * r0.H, a1 += r1.H * r0.L (m);
			/* end of inner loop second half */

		a1 = a1 >>> 15;				// scale down the signed integer
		r4 = (a0 += a1);

		// store output sample
		[i2++] = r4;


	loop_end count;
		/* end of outer loop */




	r7 = i1;
	[p0+12] = r7;				// update filter->head
	l0 = 0;						// restore all lx register by zero;
	l1 = l0;
	l2 = l0;
	l3 = l0;
	(r7:4, p5:3) = [sp++];		// restore all preserved register
	unlink;
	rts;
	.size	_fr32_interpolate, .-_fr32_interpolate
